import { $fetch } from "ohmyfetch";
import { defineDriver } from "./utils/index.mjs";
const LOG_TAG = "[unstorage] [cloudflare-http] ";
export default defineDriver((opts) => {
  if (!opts.accountId) {
    throw new Error(LOG_TAG + "`accountId` is required.");
  }
  if (!opts.namespaceId) {
    throw new Error(LOG_TAG + "`namespaceId` is required.");
  }
  let headers;
  if ("apiToken" in opts) {
    headers = { Authorization: `Bearer ${opts.apiToken}` };
  } else if ("userServiceKey" in opts) {
    headers = { "X-Auth-User-Service-Key": opts.userServiceKey };
  } else if (opts.email && opts.apiKey) {
    headers = { "X-Auth-Email": opts.email, "X-Auth-Key": opts.apiKey };
  } else {
    throw new Error(LOG_TAG + "One of the `apiToken`, `userServiceKey`, or a combination of `email` and `apiKey` is required.");
  }
  const apiURL = opts.apiURL || "https://api.cloudflare.com";
  const baseURL = `${apiURL}/client/v4/accounts/${opts.accountId}/storage/kv/namespaces/${opts.namespaceId}`;
  const kvFetch = $fetch.create({ baseURL, headers });
  const hasItem = async (key) => {
    const { success } = await kvFetch(`/values/${key}`);
    return success;
  };
  const getItem = async (key) => {
    const { result } = await kvFetch(`/values/${key}`);
    return result;
  };
  const setItem = async (key, value) => {
    return await kvFetch(`/values/${key}`, { method: "PUT", body: value });
  };
  const removeItem = async (key) => {
    return await kvFetch(`/values/${key}`, { method: "DELETE" });
  };
  const getKeys = async (base) => {
    const keys = [];
    const params = new URLSearchParams();
    if (base) {
      params.set("prefix", base);
    }
    const firstPage = await kvFetch("/keys", { params });
    firstPage.result.forEach(({ name }) => keys.push(name));
    const cursor = firstPage.result_info.cursor;
    if (cursor) {
      params.set("cursor", cursor);
    }
    while (params.has("cursor")) {
      const pageResult = await kvFetch("/keys", { params });
      pageResult.result.forEach(({ name }) => keys.push(name));
      const pageCursor = pageResult.result_info.cursor;
      if (pageCursor) {
        params.set("cursor", pageCursor);
      } else {
        params.delete("cursor");
      }
    }
    return keys;
  };
  const clear = async () => {
    const keys = await getKeys();
    const chunks = keys.reduce((acc, key, i) => {
      if (i % 1e4 === 0) {
        acc.push([]);
      }
      acc[acc.length - 1].push(key);
      return acc;
    }, [[]]);
    await Promise.all(chunks.map((chunk) => {
      return kvFetch("/bulk", {
        method: "DELETE",
        body: { keys: chunk }
      });
    }));
  };
  return {
    hasItem,
    getItem,
    setItem,
    removeItem,
    getKeys,
    clear
  };
});
